# -*- coding: utf-8 -*-
"""calc_cost.py

『SS7』『Op.Python実行』実例集
積算の計算処理を行います。

Copyright (C) 2024 UNION SYSTEM Inc.
"""


from enum import IntFlag
import os
import re
import csv
from typing import Union, List, Dict, Iterable
from Python import Ss7Python as Cmd
from Python.Ss7Python import Ss7Data, Ss7Result, ErrInfo
from ss7_oppython_utils import (
    Ss7PythonUtils,
    StartParam,
    OpenParamExchange,
    OpenParamBuckup,
    CloseParam,
    ExportResultCsvParamSave,
    ExportResultCsvParamSymbol,
    ExportResultCsvParamOutput,
)

# 各変数の定義
SS7_VERSION: str = "1.1.1.19"
CALC_ITEMS: str = "積算"
OUT_ITEMS: str = "17.積算"
NAME_CONCRETE: str = "name=延べ床面積当たりの集計表(コンクリート，型枠，鉄筋)"
NAME_STEEL: str = "name=延べ床面積当たりの集計表(鉄骨)"
CONCRETE_KEYS: List[str] = ["コンクリート体積", "型枠面積", "鉄筋重量"]
STEEL_KEYS: List[str] = ["鉄骨重量"]
CONCRETE_COLUMNS: List[str] = [
    "基礎",
    "基礎梁",
    "柱",
    "大梁",
    "片持梁",
    "小梁",
    "床",
    "耐震壁",
    "雑壁",
    "ﾊﾟﾗﾍﾟｯﾄ",
    "その他",
]
STEEL_COLUMNS: List[str] = ["柱", "大梁", "片持梁", "小梁", "鉛直ﾌﾞﾚｰｽ", "その他"]


class StructEnum(IntFlag):
    """
    構造種別の列挙型
    構造種別は複数選択可能のため、ビットフラグとして定義する
    """

    NONE = 0b00000  # 未設定
    RC = 0b00001  # RC造
    SRC = 0b00010  # SRC造
    STEEL = 0b00100  # S造
    CFT = 0b01000  # CFT造


class BuildInfo:
    """
    物件情報クラス
    """

    def __init__(
        self,
        name: str,  # 物件名
        x_span: int,  # Xスパン
        y_span: int,  # Yスパン
        n_floor: int,  # 階数
        structure: StructEnum,  # 構造種別
    ) -> None:
        self.name = name
        self.x_span = x_span
        self.y_span = y_span
        self.n_floor = n_floor
        self.structure = structure


class ResultCost:
    """
    積算結果クラス
    """

    def __init__(
        self,
        build_info: BuildInfo,  # 物件情報
    ) -> None:
        # 物件情報
        self.build_info = build_info
        # 延べ床面積
        self.total_floor_area: float = 0.0
        # コンクリート関連のデータ
        self.concrete_data: Dict[str, Dict[str, float]] = {}
        # 鉄骨関連のデータ
        self.steel_data: Dict[str, Dict[str, float]] = {}


class CalcCost:
    """
    積算計算クラス
    """

    def __init__(
        self,
    ) -> None:
        # 『Op.Python実行』の初期化
        self.ss7cmd = Ss7PythonUtils(Cmd, SS7_VERSION)
        # 『SS7』の開始
        self.ss7cmd.start(StartParam.CLEAR_LOG)

    def calc_cost(
        self,
        ikn_folders: List[str],
        result_no: int,
        result_data: Dict[str, ResultCost],
    ) -> Union[bool, Dict[str, str]]:
        """積算の計算処理を行う関数

        Args:
            ikn_folders (list): 物件フォルダ(*.ikn)のリスト
            result_no (int): 結果No.
            result_data (Dict[str, ResultCost]): 積算結果

        Returns:
            Union[bool, Dict[str, str]]: エラーが発生した場合はエラーメッセージを返す
        """
        result_no_str = f"結果{result_no}"
        for index, ikn_folder in enumerate(ikn_folders):
            # 物件データを開く(コピーして変換, バックアップは削除)
            data: Ss7Data = self.ss7cmd.open(
                ikn_folder,
                OpenParamExchange.COPY_CONVERT_AND_OPEN,
                OpenParamBuckup.DELETE_BACKUP_AND_OPEN,
            )
            if isinstance(data, ErrInfo):  # エラーでデータが取得できない場合
                return self.return_err(data)

            # 結果フォルダが存在しない場合は、計算を実行する
            result_folder_path = os.path.join(ikn_folder, f"{result_no_str}.out")
            should_calculate = not os.path.exists(result_folder_path)

            # 計算項目が計算済みかどうかを確認
            if not should_calculate:
                should_calculate = not self.ss7cmd.is_calc_end(
                    data, result_no_str, CALC_ITEMS
                )

            if should_calculate:
                calc = self.ss7cmd.calculate(data, result_no_str, CALC_ITEMS)
                if isinstance(calc, ErrInfo):  # エラーでデータが取得できない場合
                    return self.return_err(calc)

            # 結果データを取得
            result: Ss7Result = self.ss7cmd.get_result(data, result_no_str)
            if isinstance(result, ErrInfo):  # エラーでデータが取得できない場合
                return self.return_err(result)

            # 結果データが取得できた場合、結果CSVを出力する
            csv_path: str = os.path.join(ikn_folder, "cost.csv")
            # 1=上書き, 2=記号は省略しない, 1=全部材出力
            result_csv = self.ss7cmd.export_result_csv(
                result,
                OUT_ITEMS,
                csv_path.replace("/", "\\"),
                ExportResultCsvParamSave.OVERWRITE,
                ExportResultCsvParamSymbol.NOT_OMIT_SYMBOL,
                ExportResultCsvParamOutput.ALL,
            )
            if isinstance(result_csv, ErrInfo):  # エラーでデータが取得できない場合
                # 物件データを閉じて次の物件へ
                self.ss7cmd.close(
                    data,
                    CloseParam.NOT_SAVE_AND_CLOSE,
                )
                continue

            # 物件情報を取得
            # 物件名は.iknを除いたフォルダ名とする
            ikan_name = os.path.basename(ikn_folder).replace(".ikn", "")
            # 階数、Xスパン、Yスパンを取得
            build_scale = self.get_info_build_scale(ikn_folder)
            if isinstance(build_scale, ErrInfo):
                return self.return_err(build_scale)
            n_floor = build_scale["n_floor"]
            x_span = build_scale["x_span"]
            y_span = build_scale["y_span"]
            # 構造種別を取得
            structure = self.get_info_structure(ikn_folder)
            if isinstance(structure, ErrInfo):
                return self.return_err(structure)
            # 物件情報を保持
            build_info = BuildInfo(ikan_name, x_span, y_span, n_floor, structure)
            result_data[ikan_name] = ResultCost(build_info)

            # 物件データを閉じる
            self.ss7cmd.close(
                data,
                CloseParam.SAVE_AND_CLOSE,
            )

        return True

    def get_info_build_scale(self, ikn_folder: str) -> Union[Dict[str, int], ErrInfo]:
        """
        物件の階数、Xスパン、Yスパンを取得する関数

        Args:
            ikn_folder (str): 物件フォルダパス（※絶対パス）

        Returns:
            Union[Dict[str, int], ErrInfo]: 物件の階数、Xスパン、Yスパンかエラー情報
        """
        n_floor = self.ss7cmd.get_info_floor(ikn_folder)
        if isinstance(n_floor, ErrInfo):
            return n_floor
        x_span = self.ss7cmd.get_info_span_x(ikn_folder)
        if isinstance(x_span, ErrInfo):
            return x_span
        y_span = self.ss7cmd.get_info_span_y(ikn_folder)
        if isinstance(y_span, ErrInfo):
            return y_span

        return {"n_floor": n_floor, "x_span": x_span, "y_span": y_span}

    def get_info_structure(self, ikn_folder: str) -> Union[StructEnum, ErrInfo]:
        """
        物件の構造種別を取得する関数

        Args:
            ikn_folder (str): 物件フォルダパス（※絶対パス）

        Returns:
            Union[StructEnum, ErrInfo]: 構造種別かエラー情報
        """
        # RC造が含まれるか
        is_rc = self.ss7cmd.get_info_kozo_rc(ikn_folder)
        if isinstance(is_rc, ErrInfo):
            return is_rc
        # SRC造が含まれるか
        is_src = self.ss7cmd.get_info_kozo_src(ikn_folder)
        if isinstance(is_src, ErrInfo):
            return is_src
        # S造が含まれるか
        is_steel = self.ss7cmd.get_info_kozo_s(ikn_folder)
        if isinstance(is_steel, ErrInfo):
            return is_steel
        # CFT造が含まれるか
        is_cft = self.ss7cmd.get_info_kozo_cft(ikn_folder)
        if isinstance(is_cft, ErrInfo):
            return is_cft

        # 構造種別を判定
        strunture = StructEnum.NONE
        if is_rc:
            strunture |= StructEnum.RC
        if is_src:
            strunture |= StructEnum.SRC
        if is_steel:
            strunture |= StructEnum.STEEL
        if is_cft:
            strunture |= StructEnum.CFT

        return strunture

    def return_err(self, errinfo: ErrInfo) -> Dict[str, str]:
        """
        エラー情報を辞書に変換して返す関数

        Args:
            errinfo (ErrInfo): エラー情報

        Returns:
            Dict[str, str]: エラー情報（エラー番号、エラーメッセージ）
        """
        return {
            "no": str(self.ss7cmd.get_error_no(errinfo)),
            "msg": self.ss7cmd.get_error_msg(errinfo),
        }

    def arrange_result_cost(
        self, ikn_folders: List[str], result_data: Dict[str, ResultCost]
    ) -> Union[bool, Dict[str, str]]:
        """
        積算結果の整理を行う関数

        Args:
            result_data (Dict[str, ResultCost]): 積算結果

        Returns:
            Union[bool, Dict[str, str]]: エラーが発生した場合はエラーメッセージを返す
        """
        for ikan_folder in ikn_folders:
            # csvファイルのパス
            csv_path = os.path.join(ikan_folder, "cost.csv")
            if not os.path.exists(csv_path):
                continue
            # 物件名を取得
            ikan_name = os.path.basename(ikan_folder).replace(".ikn", "")
            # CSVファイルを読み取り、積算結果を整理
            read_csv(csv_path, result_data[ikan_name])
        return True


def read_csv(file_name: str, result_cost: ResultCost) -> None:
    """
    CSVファイルを読み取り、延べ床あたりの集計表のデータを取得する関数

    Args:
        file_name (str): CSVファイル名
        result_cost (ResultCost): 積算結果

    Returns:
        None
    """
    with open(file_name, mode="r", encoding="cp932") as f:
        reader: Iterable[List[str]] = csv.reader(f)
        for row in reader:
            if not row:  # 空行の場合はスキップ
                continue
            if NAME_CONCRETE in row[0]:
                read_concrete_data(reader, result_cost)
            if NAME_STEEL in row[0]:
                read_steel_data(reader, result_cost)


def get_records(row: List[str], columns: List[str]) -> Dict[str, float]:
    """
    csvファイルの行データを辞書に変換する関数

    Args:
        row (List[str]): 行データ
        columns (List[str]): 列名

    Returns:
        Dict[str, float]: 辞書データ
    """
    data: Dict[str, float] = {}
    for i, column in enumerate(columns):
        try:
            data[column] = float(row[i + 1])
        except (ValueError, IndexError):
            data[column] = 0.0
    return data


def read_concrete_data(reader: Iterable[List[str]], result_cost: ResultCost) -> None:
    """
    コンクリート関連のデータを読み取る関数

    Args:
        reader (Iterable[List[str]]): CSVリーダ
        result_cost (ResultCost): 積算結果

    Returns:
        None
    """
    concrete_keys = set(CONCRETE_KEYS)
    for row in reader:
        # 延べ床面積が0の場合は、延べ床面積を取得
        if result_cost.total_floor_area == 0.0 and "延床面積" in "".join(row):
            result_cost.total_floor_area = float(row[1].split("：")[1].split("[")[0])
        # コンクリート関連のデータを取得
        elif any(col in "".join(row) for col in concrete_keys):
            key = re.sub(r"\[.*\]|　| ", "", row[0])  # データのキーを取得
            result_cost.concrete_data[key] = get_records(row, CONCRETE_COLUMNS)
            concrete_keys.remove(key)
            # すべてのキーが処理されたら、関数から抜け出す
            if not concrete_keys:
                return


def read_steel_data(reader: Iterable[List[str]], result_cost: ResultCost) -> None:
    """
    鉄骨関連のデータを読み取る関数

    Args:
        reader (Iterable[List[str]): CSVリーダ
        result_cost (ResultCost): 積算結果

    Returns:
        None
    """
    steel_keys = set(STEEL_KEYS)
    for row in reader:
        # 延べ床面積が0の場合は、延べ床面積を取得
        if result_cost.total_floor_area == 0.0 and "延床面積" in "".join(row):
            result_cost.total_floor_area = float(row[1].split("：")[1].split("[")[0])
        # 鉄骨関連のデータを取得
        elif any(col in "".join(row) for col in steel_keys):
            key = re.sub(r"\[.*\]|　| ", "", row[0])
            result_cost.steel_data[key] = get_records(row, STEEL_KEYS)
            steel_keys.remove(key)
            # すべてのキーが処理されたら、関数から抜け出す
            if not steel_keys:
                return
