# -*- coding: utf-8 -*-
"""gui.py

『SS7』『Op.Python実行』実例集
『Op.積算』の結果一覧をグラフで表示する機能

Copyright (C) 2024 UNION SYSTEM Inc.
"""

import os
from functools import partial
import tkinter as tk
import tkinter.ttk as ttk
from tkinter import filedialog, messagebox
import japanize_matplotlib  # 日本語化用モジュール  # noqa: F401
from typing import Any, Dict, Tuple
from calc_cost import CalcCost, ResultCost
from graph import ComparePlotter, RedrawError


class CostViewerApp(tk.Frame):
    """
    積算結果比較アプリケーションクラス

    Args:
        tk (tk.Frame): TkinterのFrameクラス
    """

    def __init__(self, master: tk.Tk) -> None:
        """
        初期化

        Args:
            master (tk.Tk): TkinterのTkクラス

        Returns:
            None
        """
        super().__init__(master)
        self.result_data: Dict[str, ResultCost] = {}
        self.folder_path: str = ""  # フォルダパス
        self.result_no: int = 1  # 結果No.
        # 構造種別
        self.struct_bool: Dict[str, bool] = {
            "RC": True,
            "S": True,
            "SRC": True,
        }
        # 建築規模
        self.scale_building: Dict[str, Tuple[int, int]] = {
            "階": (1, 100),
            "Xスパン": (1, 100),
            "Yスパン": (1, 100),
        }

        # 単価（千円）
        self.unit_price: Dict[str, int] = {
            "コンクリート": 14,
            "型枠": 30,
            "鉄筋": 150,
            "鉄骨": 260,
        }

        # UI要素の定義----------------------------
        self.master: tk.Tk = master
        self.master.title("積算結果比較")  # type: ignore

        # フレーム（フォルダ選択）
        self.frame_folder = self._create_frame_folder()
        # フレーム（結果No.）
        self.frame_result_no = self._create_frame_calculate()
        # ラベルフレーム（構造種別）
        self.frame_structure = self._create_frame_structure()
        # ラベルフレーム（建築規模）
        self.frame_scale = self._create_frame_scale()
        # ラベルフレーム（単価）
        self.frame_unit_price = self._create_frame_unit_price()
        # ラベルフレーム（グラフ）
        self.frame_graph = self._create_frame_graph()
        # ラベルフレーム（詳細グラフ）
        self.frame_detail_graph = self._create_frame_detail_graph()

        # グリッド配置
        self._grid_layout()
        # グラフクラス
        self.plotter = ComparePlotter(
            self.frame_graph, self.frame_detail_graph, self.unit_price
        )

    def _create_frame_folder(self) -> tk.LabelFrame:
        """
        フォルダ選択のフレームを作成する

        Returns:
            tk.Frame: フォルダ選択のフレーム
        """
        frame_folder = tk.LabelFrame(self.master, text="物件フォルダパス")
        lbl_folder = tk.Label(frame_folder, text="フォルダ")
        self.entry_folder_path = tk.Entry(frame_folder, width=80)
        btn_refer = tk.Button(frame_folder, text="参照", command=self.on_refer)

        # グリッド配置
        grid_options = {"padx": 5, "pady": 5, "row": 0}
        grid_with_options = partial(tk.Widget.grid, **grid_options)
        grid_with_options(lbl_folder, column=0)
        grid_with_options(self.entry_folder_path, column=1)
        grid_with_options(btn_refer, column=2)

        return frame_folder

    def _create_frame_calculate(self) -> tk.LabelFrame:
        """
        計算実行のフレームを作成する

        Returns:
            tk.Frame: 計算実行のフレーム
        """
        frame_result_no = tk.LabelFrame(self.master, text="計算実行")

        lbl_result_no = tk.Label(frame_result_no, text="結果No.")
        self.cmb_result_no = ttk.Combobox(
            frame_result_no,
            values=["1", "2", "3", "4", "5"],
            state="readonly",
            width=5,
        )
        self.cmb_result_no.set("1")
        self.cmb_result_no.bind("<<ComboboxSelected>>", self.on_select_result_no)
        btn_run = tk.Button(frame_result_no, text="実行", command=self.on_run)

        # グリッド配置
        grid_options = {"padx": 5, "pady": 5, "row": 0}
        grid_with_options = partial(tk.Widget.grid, **grid_options)
        grid_with_options(lbl_result_no, column=0)
        grid_with_options(self.cmb_result_no, column=1)
        grid_with_options(btn_run, column=2)

        return frame_result_no

    def _create_frame_unit_price(self) -> tk.LabelFrame:
        """
        単価のラベルフレームを作成する

        Returns:
            tk.LabelFrame: 単価のラベルフレーム
        """

        def on_validate(text: str) -> bool:
            """入力値の検証を行う関数

            Args:
                text (str): 入力値

            Returns:
                bool: 入力値が正の整数の場合はTrue、それ以外はFalse
            """
            return text.isdigit() and int(text) >= 0

        frame_unit_price = tk.LabelFrame(self.master, text="単価")

        lbl_up_concrete = tk.Label(frame_unit_price, text="コンクリート")
        lbl_up_formwork = tk.Label(frame_unit_price, text="型枠")
        lbl_up_reinforcement = tk.Label(frame_unit_price, text="鉄筋")
        lbl_up_steel = tk.Label(frame_unit_price, text="鉄骨")
        lbl_up_volume = tk.Label(frame_unit_price, text="[千円/m3]")
        lbl_up_area = tk.Label(frame_unit_price, text="[千円/m2]")
        lbl_up_weight1 = tk.Label(frame_unit_price, text="[千円/t]")
        lbl_up_weight2 = tk.Label(frame_unit_price, text="[千円/t]")

        # バリデーションを設定
        vcmd = (self.master.register(on_validate), "%S")

        self.entry_up_concrete = tk.Entry(
            frame_unit_price,
            width=10,
            justify=tk.RIGHT,
            validate="key",
            validatecommand=vcmd,
        )
        self.entry_up_formwork = tk.Entry(
            frame_unit_price,
            width=10,
            justify=tk.RIGHT,
            validate="key",
            validatecommand=vcmd,
        )
        self.entry_up_reinforcement = tk.Entry(
            frame_unit_price,
            width=10,
            justify=tk.RIGHT,
            validate="key",
            validatecommand=vcmd,
        )
        self.entry_up_steel = tk.Entry(
            frame_unit_price,
            width=10,
            justify=tk.RIGHT,
            validate="key",
            validatecommand=vcmd,
        )

        self.entry_up_concrete.bind("<KeyRelease>", self.on_change_unit_price)
        self.entry_up_formwork.bind("<KeyRelease>", self.on_change_unit_price)
        self.entry_up_reinforcement.bind("<KeyRelease>", self.on_change_unit_price)
        self.entry_up_steel.bind("<KeyRelease>", self.on_change_unit_price)

        # デフォルト値を設定
        self.entry_up_concrete.insert(0, str(self.unit_price["コンクリート"]))
        self.entry_up_formwork.insert(0, str(self.unit_price["型枠"]))
        self.entry_up_reinforcement.insert(0, str(self.unit_price["鉄筋"]))
        self.entry_up_steel.insert(0, str(self.unit_price["鉄骨"]))

        # グリッド配置
        grid_options = {"padx": 5, "pady": 5, "row": 0}
        grid_with_options = partial(tk.Widget.grid, **grid_options)

        grid_with_options(lbl_up_concrete, column=0, sticky=tk.E)
        grid_with_options(self.entry_up_concrete, column=1)
        grid_with_options(lbl_up_volume, column=2, sticky=tk.W)
        grid_with_options(lbl_up_formwork, column=3, sticky=tk.E)
        grid_with_options(self.entry_up_formwork, column=4)
        grid_with_options(lbl_up_area, column=5, sticky=tk.W)
        grid_with_options(lbl_up_reinforcement, column=6, sticky=tk.E)
        grid_with_options(self.entry_up_reinforcement, column=7)
        grid_with_options(lbl_up_weight1, column=8, sticky=tk.W)
        grid_with_options(lbl_up_steel, column=9, sticky=tk.E)
        grid_with_options(self.entry_up_steel, column=10)
        grid_with_options(lbl_up_weight2, column=11, sticky=tk.W)

        return frame_unit_price

    def _create_frame_structure(self) -> tk.LabelFrame:
        """
        構造種別のラベルフレームを作成する

        Returns:
            tk.LabelFrame: 構造種別のラベルフレーム
        """
        frame_structure = tk.LabelFrame(self.master, text="構造種別")
        # チェックボックス
        self.var_rc = tk.BooleanVar(value=self.struct_bool["RC"])
        self.var_s = tk.BooleanVar(value=self.struct_bool["S"])
        self.var_src = tk.BooleanVar(value=self.struct_bool["SRC"])
        self.chk_rc = tk.Checkbutton(
            frame_structure,
            text="RC造",
            variable=self.var_rc,
            command=self.on_checkbox_click,
        )
        self.chk_s = tk.Checkbutton(
            frame_structure,
            text="S造",
            variable=self.var_s,
            command=self.on_checkbox_click,
        )
        self.chk_src = tk.Checkbutton(
            frame_structure,
            text="SRC造",
            variable=self.var_src,
            command=self.on_checkbox_click,
        )

        # グリッド配置
        grid_options = {"padx": 5, "pady": 5, "row": 0}
        grid_with_options = partial(tk.Widget.grid, **grid_options)

        grid_with_options(self.chk_rc, column=0)
        grid_with_options(self.chk_s, column=1)
        grid_with_options(self.chk_src, column=2)

        return frame_structure

    def _create_frame_scale(self) -> tk.LabelFrame:
        """
        建築規模のラベルフレームを作成する

        Returns:
            tk.LabelFrame: 建築規模のラベルフレーム
        """
        frame_scale = tk.LabelFrame(self.master, text="建築規模")
        lbl_floor = tk.Label(frame_scale, text="階")
        lbl_x_span = tk.Label(frame_scale, text="Xスパン")
        lbl_y_span = tk.Label(frame_scale, text="Yスパン")

        # 上下限のコンボボックス
        scale_values = [str(i) for i in range(1, 101)]
        self.cmb_scale_floor_lower = ttk.Combobox(
            frame_scale, values=scale_values, state="readonly", width=5
        )
        self.cmb_scale_floor_upper = ttk.Combobox(
            frame_scale, values=scale_values, state="readonly", width=5
        )
        self.cmb_scale_x_span_lower = ttk.Combobox(
            frame_scale, values=scale_values, state="readonly", width=5
        )
        self.cmb_scale_x_span_upper = ttk.Combobox(
            frame_scale, values=scale_values, state="readonly", width=5
        )
        self.cmb_scale_y_span_lower = ttk.Combobox(
            frame_scale, values=scale_values, state="readonly", width=5
        )
        self.cmb_scale_y_span_upper = ttk.Combobox(
            frame_scale, values=scale_values, state="readonly", width=5
        )

        # 初期値の設定
        self.cmb_scale_floor_lower.set(self.scale_building["階"][0])
        self.cmb_scale_floor_upper.set(self.scale_building["階"][1])
        self.cmb_scale_x_span_lower.set(self.scale_building["Xスパン"][0])
        self.cmb_scale_x_span_upper.set(self.scale_building["Xスパン"][1])
        self.cmb_scale_y_span_lower.set(self.scale_building["Yスパン"][0])
        self.cmb_scale_y_span_upper.set(self.scale_building["Yスパン"][1])

        # 関数のバインド
        self.cmb_scale_floor_lower.bind("<<ComboboxSelected>>", self.on_change_scale)
        self.cmb_scale_floor_upper.bind("<<ComboboxSelected>>", self.on_change_scale)
        self.cmb_scale_x_span_lower.bind("<<ComboboxSelected>>", self.on_change_scale)
        self.cmb_scale_x_span_upper.bind("<<ComboboxSelected>>", self.on_change_scale)
        self.cmb_scale_y_span_lower.bind("<<ComboboxSelected>>", self.on_change_scale)
        self.cmb_scale_y_span_upper.bind("<<ComboboxSelected>>", self.on_change_scale)

        # グリッド配置
        grid_options = {"padx": 5, "pady": 5, "row": 0}
        grid_with_options = partial(tk.Widget.grid, **grid_options)
        grid_with_options(lbl_floor, column=0)
        grid_with_options(self.cmb_scale_floor_lower, column=1)
        grid_with_options(self.cmb_scale_floor_upper, column=2)
        grid_with_options(lbl_x_span, column=3)
        grid_with_options(self.cmb_scale_x_span_lower, column=4)
        grid_with_options(self.cmb_scale_x_span_upper, column=5)
        grid_with_options(lbl_y_span, column=6)
        grid_with_options(self.cmb_scale_y_span_lower, column=7)
        grid_with_options(self.cmb_scale_y_span_upper, column=8)

        return frame_scale

    def _create_frame_graph(self) -> tk.LabelFrame:
        """
        グラフのラベルフレームを作成する

        Returns:
            tk.LabelFrame: グラフのラベルフレーム
        """
        frame_graph = tk.LabelFrame(self.master, text="比較グラフ")
        frame_graph.config(width=400, height=300)

        return frame_graph

    def _create_frame_detail_graph(self) -> tk.LabelFrame:
        """
        詳細グラフのラベルフレームを作成する

        Returns:
            tk.LabelFrame: 詳細グラフのラベルフレーム
        """
        frame_detail_graph = tk.LabelFrame(self.master, text="詳細グラフ")
        frame_detail_graph.config(width=400, height=300)

        return frame_detail_graph

    def _grid_layout(self) -> None:
        """
        グリッド配置を行う
        """
        grid_options = {"padx": 5, "pady": 2, "sticky": tk.NW}
        grid_with_options = partial(tk.Widget.grid, **grid_options)
        grid_with_options(self.frame_folder, row=0, column=0, columnspan=4)
        grid_with_options(self.frame_result_no, row=1, column=0, columnspan=4)
        grid_with_options(self.frame_unit_price, row=2, column=0, columnspan=4)
        grid_with_options(self.frame_structure, row=3, column=0, columnspan=4)
        grid_with_options(self.frame_scale, row=4, column=0, columnspan=4)
        grid_with_options(self.frame_graph, row=5, column=0, columnspan=2)
        grid_with_options(self.frame_detail_graph, row=5, column=2, columnspan=2)

    def on_refer(self) -> None:
        """
        参照ボタン押下時の処理
        """
        self.folder_path = filedialog.askdirectory()
        self.entry_folder_path.insert(0, self.folder_path)

    def on_select_result_no(self, event: Any) -> None:
        """
        結果No.のコンボボックス選択時の処理
        """
        self.result_no = int(self.cmb_result_no.get())

    def on_checkbox_click(self) -> None:
        """
        チェックボックスクリック時の処理
        """
        try:
            bool_rc = self.var_rc.get()
            bool_s = self.var_s.get()
            bool_src = self.var_src.get()

            temp_struct_bool = self.struct_bool.copy()
            temp_struct_bool["RC"] = bool_rc
            temp_struct_bool["S"] = bool_s
            temp_struct_bool["SRC"] = bool_src

            # グラフの再描画
            if self.result_data:
                if not self.plotter.redraw_plot(
                    self.result_data,
                    self.scale_building,
                    temp_struct_bool,
                    self.unit_price,
                ):
                    raise RedrawError

            # 構造種別の更新
            self.struct_bool["RC"] = bool_rc
            self.struct_bool["S"] = bool_s
            self.struct_bool["SRC"] = bool_src

        except RedrawError:
            # 入力をキャンセルして元の値に戻す
            self.var_rc.set(self.struct_bool["RC"])
            self.var_s.set(self.struct_bool["S"])
            self.var_src.set(self.struct_bool["SRC"])

    def on_change_scale(self, event: Any) -> None:
        """
        規模コンボボックス変更時の処理
        """
        try:
            floor_lower = int(self.cmb_scale_floor_lower.get())
            floor_upper = int(self.cmb_scale_floor_upper.get())
            x_span_lower = int(self.cmb_scale_x_span_lower.get())
            x_span_upper = int(self.cmb_scale_x_span_upper.get())
            y_span_lower = int(self.cmb_scale_y_span_lower.get())
            y_span_upper = int(self.cmb_scale_y_span_upper.get())

            if (
                floor_lower > floor_upper
                or x_span_lower > x_span_upper
                or y_span_lower > y_span_upper
            ):
                raise RedrawError

            temp_scale_building = self.scale_building.copy()
            temp_scale_building["階"] = (floor_lower, floor_upper)
            temp_scale_building["Xスパン"] = (x_span_lower, x_span_upper)
            temp_scale_building["Yスパン"] = (y_span_lower, y_span_upper)

            # グラフの再描画
            if self.result_data:
                if not self.plotter.redraw_plot(
                    self.result_data,
                    temp_scale_building,
                    self.struct_bool,
                    self.unit_price,
                ):
                    raise RedrawError

            # スケールの更新
            self.scale_building["階"] = (floor_lower, floor_upper)
            self.scale_building["Xスパン"] = (x_span_lower, x_span_upper)
            self.scale_building["Yスパン"] = (y_span_lower, y_span_upper)

        except RedrawError:
            # 入力をキャンセルして元の値に戻す
            self.cmb_scale_floor_lower.set(self.scale_building["階"][0])
            self.cmb_scale_floor_upper.set(self.scale_building["階"][1])
            self.cmb_scale_x_span_lower.set(self.scale_building["Xスパン"][0])
            self.cmb_scale_x_span_upper.set(self.scale_building["Xスパン"][1])
            self.cmb_scale_y_span_lower.set(self.scale_building["Yスパン"][0])
            self.cmb_scale_y_span_upper.set(self.scale_building["Yスパン"][1])

    def on_change_unit_price(self, event: Any) -> None:
        """
        単価テキストボックス変更時の処理
        """
        self.unit_price["コンクリート"] = int(self.entry_up_concrete.get())
        self.unit_price["型枠"] = int(self.entry_up_formwork.get())
        self.unit_price["鉄筋"] = int(self.entry_up_reinforcement.get())
        self.unit_price["鉄骨"] = int(self.entry_up_steel.get())

        # グラフの再描画
        if self.result_data:
            self.plotter.redraw_plot(
                self.result_data, self.scale_building, self.struct_bool, self.unit_price
            )

    def on_run(self) -> None:
        """
        実行ボタン押下時の処理
        """
        if not self.is_selected_folder():
            messagebox.showerror(
                title="エラー", message="フォルダが選択されていません。"
            )
            return
        if self.ask_run():

            # SS7Cmdのセットアップ
            calcost = CalcCost()

            # 物件フォルダ(*.ikn)の一覧を取得
            ikn_folders: list = self.get_ikn_dir(app.folder_path)
            # 積算の計算実行
            process = calcost.calc_cost(ikn_folders, app.result_no, self.result_data)

            if process is True:
                messagebox.showinfo(
                    title="計算実行",
                    message="計算が完了しました。",
                )
            else:
                err_no, err_msg = process["no"], process["msg"]
                self.show_error(err_no, err_msg)

            # 積算結果の整理
            process = calcost.arrange_result_cost(ikn_folders, self.result_data)
            if process is True:
                messagebox.showinfo(
                    title="積算結果整理",
                    message="積算結果の整理が完了しました。",
                )
                self.plotter.plot_compare_graph(self.result_data)
            else:
                err_no, err_msg = process["no"], process["msg"]
                self.show_error(err_no, err_msg)

    def is_selected_folder(self) -> bool:
        """
        フォルダが選択されているか
        """
        return bool(self.folder_path)

    def get_ikn_dir(self, folder_path: str) -> list:
        """指定フォルダから物件フォルダ(*.ikn)の一覧を取得し、ディレクトリの一覧を返す関数

        Args:
            folder_path (str): 指定フォルダ

        Returns:
            ikn_dir (list): 物件フォルダ(*.ikn)の一覧
        """
        ikn_dir: list = []
        for file in os.listdir(folder_path):
            if file.endswith(".ikn"):
                ikn_dir.append(os.path.join(folder_path, file))
        return ikn_dir

    def ask_run(self) -> bool:
        """
        実行確認ダイアログ表示
        """
        ret = messagebox.askokcancel(
            title="計算実行",
            message=(
                f"結果{self.result_no}.out の結果フォルダは上書きされます。\
                \n\n実行しますか？\n\n※ロックされた結果フォルダは計算対象外です。"
            ),
        )
        return ret

    def show_error(self, err_no: int, err_msg: str) -> None:
        """エラーメッセージを表示する

        Args:
            err_no (int): エラー番号
            err_msg (str): エラーメッセージ

        Returns:
            None

        Note:
            エラーメッセージを表示し、計算処理を中断します。
        """
        messagebox.showerror(
            title="エラー",
            message=f"エラーが発生しました。\n\nエラー番号：{err_no}\nエラーメッセージ：{err_msg}",
        )


if __name__ == "__main__":
    root = tk.Tk()
    app = CostViewerApp(master=root)
    app.mainloop()
