# -*- coding: utf-8 -*-
"""gui.py

『SS7』『Op.Python実行』実例集
比較グラフと詳細グラフを描画するモジュール

Copyright (C) 2024 UNION SYSTEM Inc.
"""

from typing import Dict, List, Tuple, Optional
import tkinter as tk
from tkinter import messagebox
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from matplotlib.offsetbox import AnnotationBbox, TextArea
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.backend_bases import Event
import japanize_matplotlib  # 日本語化用モジュール  # noqa: F401
from calc_cost import ResultCost, StructEnum, CONCRETE_KEYS, STEEL_KEYS


class RedrawError(Exception):
    """グラフの再描画に失敗した場合の例外"""

    pass


class ComparePlotter:
    """
    積算結果の比較グラフを描画するクラス
    """

    # 構造種別に応じた色の辞書
    COLOR_DICT = {
        StructEnum.RC: "red",
        StructEnum.STEEL: "blue",
        StructEnum.SRC: "green",
        StructEnum.CFT: "yellow",
        StructEnum.RC | StructEnum.STEEL: "purple",
        StructEnum.RC | StructEnum.SRC: "orange",
        StructEnum.RC | StructEnum.CFT: "brown",
        StructEnum.STEEL | StructEnum.SRC: "pink",
        StructEnum.STEEL | StructEnum.CFT: "gray",
        StructEnum.SRC | StructEnum.CFT: "cyan",
        StructEnum.RC | StructEnum.STEEL | StructEnum.SRC: "magenta",
        StructEnum.RC | StructEnum.STEEL | StructEnum.CFT: "olive",
        StructEnum.RC | StructEnum.SRC | StructEnum.CFT: "lime",
        StructEnum.STEEL | StructEnum.SRC | StructEnum.CFT: "teal",
        StructEnum.RC | StructEnum.STEEL | StructEnum.SRC | StructEnum.CFT: "navy",
    }

    # 構造種別に応じたラベルの辞書
    LABEL_DICT = {
        StructEnum.RC: "RC",
        StructEnum.STEEL: "S",
        StructEnum.SRC: "SRC",
        StructEnum.CFT: "CFT",
        StructEnum.RC | StructEnum.STEEL: "RC + S",
        StructEnum.RC | StructEnum.SRC: "RC + SRC",
        StructEnum.RC | StructEnum.CFT: "RC + CFT",
        StructEnum.STEEL | StructEnum.SRC: "S + SRC",
        StructEnum.STEEL | StructEnum.CFT: "S + CFT",
        StructEnum.SRC | StructEnum.CFT: "SRC + CFT",
        StructEnum.RC | StructEnum.STEEL | StructEnum.SRC: "RC + S + SRC",
        StructEnum.RC | StructEnum.STEEL | StructEnum.CFT: "RC + S + CFT",
        StructEnum.RC | StructEnum.SRC | StructEnum.CFT: "RC + SRC + CFT",
        StructEnum.STEEL | StructEnum.SRC | StructEnum.CFT: "S + SRC + CFT",
        StructEnum.RC
        | StructEnum.STEEL
        | StructEnum.SRC
        | StructEnum.CFT: "RC + S + SRC + CFT",
    }

    def __init__(
        self, master: tk.Misc, master_detail: tk.Misc, unit_price: Dict[str, int]
    ) -> None:
        """
        初期化

        Args:
            master (tk.Misc): グラフを配置するTkinterウィジェット
            master_detail (tk.Misc): 詳細グラフを配置するTkinterウィジェット
            unit_price (Dict[str, int]): 単価

        Returns:
            None
        """
        self.master = master
        self.master_detail = master_detail
        self.unit_price: Dict[str, int] = unit_price
        self.annotation: Optional[AnnotationBbox] = None
        self.fig: Optional[Figure] = None
        self.ax: Optional[Axes] = None
        self.result_data_for_view: Dict[str, ResultCost] = {}
        self.area_weight_data: Dict[str, Dict[str, float]] = {}
        self.x_range: float = 0
        self.y_range: float = 0
        self.category_keys = {
            "コンクリート体積": "コンクリート",
            "型枠面積": "型枠",
            "鉄筋重量": "鉄筋",
            "鉄骨重量": "鉄骨",
        }
        self.detail_graph = DeteilGraph(self.master_detail)
        self.canvas: Optional[FigureCanvasTkAgg] = None

    def plot_compare_graph(self, result_data: Dict[str, ResultCost]) -> None:
        """
        比較グラフを描画する

        Args:
            calcost (CalcCost): 積算計算オブジェクト
            result_data (Dict[str, ResultCost]): 積算結果データ
            unit_price (Dict[str, int]): 単価

        Returns:
            None
        """
        self.area_weight_data = self._prepare_data(result_data)
        self._create_figure()
        self._plot_data(self.area_weight_data)
        self._set_labels_and_legend()
        self._add_hover_event()
        self._add_click_event()
        self._display_graph()

    def _prepare_data(
        self,
        result_data: Dict[str, ResultCost],
    ) -> Dict[str, Dict[str, float]]:
        """
        結果データを描画用に準備する

        Args:
            result_data (Dict[str, ResultCost]): 積算結果データ

        Returns:
            Dict[str, Dict[str, float]]: 描画用データ
        """
        self.result_data_for_view = result_data
        area_weight_data: Dict[str, Dict[str, float]] = {}
        for building_name, result in result_data.items():
            area_weight_data[building_name] = {
                "total_floor_area": result.total_floor_area,
                "total_weight": self._calculate_total_cost(result, self.unit_price),
                "structure": result.build_info.structure,
            }
        return area_weight_data

    def _calculate_total_cost(
        self, result: ResultCost, unit_price: Dict[str, int]
    ) -> float:
        """
        積算結果の総コストを計算する

        Args:
            result (ResultCost): 積算結果データ
            unit_price (Dict[str, int]): 単価

        Returns:
            float: 総コスト
        """
        total_cost: float = 0.0

        # 合計金額計算
        for key in CONCRETE_KEYS:
            data = result.concrete_data.get(key)
            if data is not None:
                total_cost += sum(data.values()) * unit_price[self.category_keys[key]]
        for key in STEEL_KEYS:
            data = result.steel_data.get(key)
            if data is not None:
                total_cost += sum(data.values()) * unit_price[self.category_keys[key]]

        return total_cost

    def _create_figure(self) -> None:
        """
        グラフの図を作成する

        Returns:
            None
        """
        self.fig = Figure(figsize=(7, 4), dpi=100)
        self.ax = self.fig.add_subplot(111)

    def _plot_data(self, area_weight_data: Dict[str, Dict[str, float]]) -> None:
        """
        データをプロットする

        Args:
            area_weight_data (Dict[str, Dict[str, float]]): 描画用データ

        Returns:
            None
        """
        if self.ax is None:
            return

        for struct in self.COLOR_DICT.keys():
            x, y = self._get_data_for_structure(area_weight_data, struct)
            if x and y:
                self.ax.scatter(
                    x,
                    y,
                    label=self.LABEL_DICT[struct],
                    color=self.COLOR_DICT[struct],
                    marker="o",
                )

        # グラフ描画領域の上限下限を取得
        x_limits = self.ax.get_xlim()
        y_limits = self.ax.get_ylim()

        # データの範囲を計算
        self.x_range = x_limits[1] - x_limits[0]
        self.y_range = y_limits[1] - y_limits[0]

    def _get_data_for_structure(
        self, area_weight_data: Dict[str, Dict[str, float]], struct: StructEnum
    ) -> Tuple[List[float], List[float]]:
        """
        特定の構造種別のデータを取得する

        Args:
            area_weight_data (Dict[str, Dict[str, float]]): 描画用データ
            struct (StructEnum): 構造種別

        Returns:
            Tuple[List[float], List[float]]: X: 延べ床面積のリスト、Y: 総コストのリスト
        """
        x = []
        y = []
        for building_data in area_weight_data.values():
            if building_data["structure"] == struct:
                x.append(building_data["total_floor_area"])
                y.append(building_data["total_weight"])
        return x, y

    def _set_labels_and_legend(self) -> None:
        """
        ラベルと凡例を設定する

        Returns:
            None
        """
        if self.ax is None or self.fig is None:
            return
        self.ax.set_xlabel("延べ床面積[m2]")
        self.ax.set_ylabel("総コスト[千円/m2]")
        self.ax.legend(loc="center left", bbox_to_anchor=(1, 0.5), fontsize="small")
        self.fig.tight_layout()

    def _add_hover_event(self) -> None:
        """
        ホバーイベントを追加する

        Returns:
            None
        """
        if self.fig is None:
            return
        self.fig.canvas.mpl_connect("motion_notify_event", self._on_hover)

    def _add_click_event(self) -> None:
        """
        クリックイベントを追加する

        Returns:
            None
        """
        if self.fig is None:
            return
        self.fig.canvas.mpl_connect("button_press_event", self._on_click)

    def _on_hover(self, event: Event) -> None:
        """
        ホバー時の処理

        Args:
            event (Event): イベント情報

        Returns:
            None
        """
        if (
            not hasattr(event, "xdata")
            or not hasattr(event, "ydata")
            or event.xdata is None
            or event.ydata is None
        ):
            self._remove_annotation()
            return

        for building_name, data in self.area_weight_data.items():
            x, y = data["total_floor_area"], data["total_weight"]
            if self._is_point_hovered(event, x, y):
                self._show_annotation(building_name, x, y)
                break
        else:
            self._remove_annotation()

    def _is_point_hovered(self, event: Event, x: float, y: float) -> bool:
        """
        プロットにホバーされているかどうかを判定する

        Args:
            event (Event): イベント情報
            x (float): 延べ床面積
            y (float): 総コスト

        Returns:
            bool: ホバーされている場合はTrue、それ以外はFalse
        """
        if (
            not hasattr(event, "xdata")
            or not hasattr(event, "ydata")
            or event.xdata is None
            or event.ydata is None
        ):
            return False

        # ホバーされている点との距離を計算
        dx = abs(event.xdata - x) / self.x_range
        dy = abs(event.ydata - y) / self.y_range

        # 距離の閾値（この値は調整可能）
        threshold = 0.02

        return dx < threshold and dy < threshold

    def _show_annotation(self, building_name: str, x: float, y: float) -> None:
        """
        アノテーションを表示する

        Args:
            building_name (str): 物件名
            x (float): 延べ床面積
            y (float): 総コスト

        Returns:
            None
        """
        if self.ax is None or self.fig is None:
            return
        self._remove_annotation()
        # 物件名、階数、Xスパン、Yスパンを表示するテキストを作成
        text = (
            f"物件名：{building_name}\n"
            f"階数: {self.result_data_for_view[building_name].build_info.n_floor}階\n"
            f"Xスパン: {self.result_data_for_view[building_name].build_info.x_span}\n"
            f"Yスパン: {self.result_data_for_view[building_name].build_info.y_span}"
        )
        offsetbox = TextArea(text)
        self.annotation = AnnotationBbox(
            offsetbox,
            (x, y),
            xybox=(-30, -30),
            xycoords="data",
            boxcoords="offset points",
            arrowprops=dict(arrowstyle="->"),
            bboxprops=dict(
                boxstyle="round,pad=0.5", edgecolor="black", facecolor="white"
            ),
        )
        self.ax.add_artist(self.annotation)
        self.fig.canvas.draw_idle()

    def _remove_annotation(self) -> None:
        """
        アノテーションを削除する

        Returns:
            None
        """
        if self.fig is None or self.annotation is None:
            return
        if self.annotation:
            self.annotation.remove()
            self.annotation = None
            self.fig.canvas.draw_idle()

    def _display_graph(self) -> None:
        """
        グラフを表示する

        Returns:
            None
        """
        if self.fig is None:
            return
        if self.canvas is not None:
            self.canvas.get_tk_widget().destroy()
        self.canvas = FigureCanvasTkAgg(self.fig, master=self.master)
        self.canvas.draw()
        self.canvas.get_tk_widget().pack(fill="both", expand=True)

    def _on_click(self, event: Event) -> None:
        """
        クリック時の処理

        Args:
            event (Event): イベント情報

        Returns:
            None
        """
        if (
            not hasattr(event, "xdata")
            or not hasattr(event, "ydata")
            or event.xdata is None
            or event.ydata is None
        ):
            return

        # クリックされた点に対応する詳細グラフを表示
        for building_name, data in self.area_weight_data.items():
            x, y = data["total_floor_area"], data["total_weight"]
            if self._is_point_hovered(event, x, y):
                self._show_detail_graph(building_name)
                break

    def _show_detail_graph(self, building_name: str) -> None:
        """
        詳細グラフを表示する

        Args:
            building_name (str): 物件名

        Returns:
            None
        """
        self.detail_graph.plot_deteil_graph(
            self.result_data_for_view[building_name],
            self.unit_price,
            self.category_keys,
        )

    def redraw_plot(
        self,
        result_data: Dict[str, ResultCost],
        scale_building: Dict[str, Tuple[int, int]],
        struct_bool: Dict[str, bool],
        unit_price: Dict[str, int],
    ) -> bool:
        """
        積算結果の比較グラフを再描画する

        Args:
            result_data (Dict[str, ResultCost]): 積算結果データ
            scale_building (Dict[str, Tuple[int, int]]): 建築規模
            struct_bool (Dict[str, bool]): 構造種別
            unit_price (Dict[str, int]): 単価

        Returns:
            bool: 再描画が成功した場合はTrue、それ以外はFalse
        """
        # 建築規模に基づいてフィルタリング
        floor_lower = scale_building["階"][0]
        floor_upper = scale_building["階"][1]
        x_span_lower = scale_building["Xスパン"][0]
        x_span_upper = scale_building["Xスパン"][1]
        y_span_lower = scale_building["Yスパン"][0]
        y_span_upper = scale_building["Yスパン"][1]

        filtered_data: Dict[str, ResultCost] = {
            building_name: data
            for building_name, data in result_data.items()
            if all(
                [
                    floor_lower <= data.build_info.n_floor <= floor_upper,
                    x_span_lower <= data.build_info.x_span <= x_span_upper,
                    y_span_lower <= data.build_info.y_span <= y_span_upper,
                ]
            )
        }

        if not filtered_data:
            messagebox.showerror(title="エラー", message="描画データがありません。")
            return False

        # 構造種別に基づいてフィルタリング
        rc = struct_bool["RC"]
        s = struct_bool["S"]
        src = struct_bool["SRC"]

        selected_structs = StructEnum.NONE
        if rc:
            selected_structs |= StructEnum.RC
        if s:
            selected_structs |= StructEnum.STEEL
        if src:
            selected_structs |= StructEnum.SRC

        filtered_data = {
            building_name: data
            for building_name, data in filtered_data.items()
            if data.build_info.structure & selected_structs
        }

        if not filtered_data:
            messagebox.showerror(title="エラー", message="描画データがありません。")
            return False

        # 単価を更新
        self.unit_price = unit_price

        # グラフを初期化
        if self.fig is not None:
            self.fig.clear()

        # グラフを再描画
        self.plot_compare_graph(filtered_data)

        return True


class DeteilGraph:
    """
    積算結果の詳細グラフを描画するクラス
    """

    def __init__(self, master: tk.Misc) -> None:
        """
        初期化メソッド

        Args:
            master (tk.Misc): グラフを配置するTkinterウィジェット

        Returns:
            None
        """
        self.master = master
        self.fig: Optional[Figure] = None
        self.ax: Optional[Axes] = None
        self.canvas: Optional[FigureCanvasTkAgg] = None

    def plot_deteil_graph(
        self,
        result_cost: ResultCost,
        unit_price: Dict[str, int],
        category_keys: Dict[str, str],
    ) -> None:
        """
        詳細グラフを描画する

        Args:
            result_cost (ResultCost): 積算結果データ
            unit_price (Dict[str, int]): 単価
            category_keys (Dict[str, str]): カテゴリーキー

        Returns:
            None
        """
        self.fig = None  # グラフを初期化
        self._create_figure()
        concrete_totals, steel_totals = self._calculate_totals(
            result_cost, unit_price, category_keys
        )
        self._plot_data(concrete_totals, steel_totals)
        self._set_labels_and_legend()
        self._display_graph()

    def _create_figure(self) -> None:
        """
        グラフの図を作成する

        Returns:
            None
        """
        self.fig = Figure(figsize=(5, 4), dpi=100)
        self.ax = self.fig.add_subplot(111)

    def _calculate_totals(
        self,
        result_cost: ResultCost,
        unit_price: Dict[str, int],
        category_keys: Dict[str, str],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """
        各カテゴリーの総コストを計算する

        Args:
            result_cost (ResultCost): 積算結果データ
            unit_price (Dict[str, int]): 単価
            category_keys (Dict[str, str]): カテゴリーキー

        Returns:
            Tuple[Dict[str, float], Dict[str, float]]: コンクリート関連の累積、鉄骨関連の累積
        """
        concrete_totals = {key: 0.0 for key in CONCRETE_KEYS}
        steel_totals = {key: 0.0 for key in STEEL_KEYS}

        # コンクリート関連の累積を計算
        for key in CONCRETE_KEYS:
            data = result_cost.concrete_data.get(key)
            if data is not None:
                concrete_totals[key] = (
                    sum(data.values()) * unit_price[category_keys[key]]
                )

        # 鉄骨関連の累積を計算
        for key in STEEL_KEYS:
            data = result_cost.steel_data.get(key)
            if data is not None:
                steel_totals[key] = sum(data.values()) * unit_price[category_keys[key]]

        return concrete_totals, steel_totals

    def _plot_data(
        self, concrete_totals: Dict[str, float], steel_totals: Dict[str, float]
    ) -> None:
        """
        データをプロットする

        Args:
            concrete_totals (Dict[str, float]): コンクリート関連の総コスト
            steel_totals (Dict[str, float]): 鉄骨関連の総コスト

        Returns:
            None
        """
        if self.ax is None:
            return

        # コンクリート関連のデータをプロット
        concrete_keys = ["コンクリート", "型枠", "鉄筋"]
        concrete_values = list(concrete_totals.values())
        self.ax.bar(
            concrete_keys, concrete_values, label="コンクリート関連", color="blue"
        )

        # 鉄骨関連のデータをプロット
        steel_keys = ["鉄骨"]
        steel_values = list(steel_totals.values())
        self.ax.bar(
            steel_keys,
            steel_values,
            label="鉄骨関連",
            color="red",
        )

    def _set_labels_and_legend(self) -> None:
        """
        ラベルと凡例を設定する

        Returns:
            None
        """
        if self.ax is None or self.fig is None:
            return
        self.ax.set_ylabel("コスト[千円/m2]")
        self.ax.legend(loc="center left", bbox_to_anchor=(1, 0.5), fontsize="small")
        self.fig.tight_layout()

    def _display_graph(self) -> None:
        """
        グラフを表示する

        Returns:
            None
        """
        if self.fig is None:
            return
        if self.canvas is not None:
            self.canvas.get_tk_widget().destroy()
        self.canvas = FigureCanvasTkAgg(self.fig, master=self.master)
        self.canvas.draw()
        self.canvas.get_tk_widget().pack(fill="both", expand=True)
