# -*- coding: utf-8 -*-
"""ss7_oppython_utils.py

『SS7』『Op.Python実行』実例集
『Op.Python実行』の共通モジュール

Copyright (C) 2024 UNION SYSTEM Inc.
"""

from enum import Enum
from typing import Optional, Union
from Python import Ss7Python as Cmd
from Python.Ss7Python import Ss7Data, Ss7Result, ErrInfo


class StartParam(Enum):
    """Start関数のパラメータ"""

    CLEAR_LOG = 1  # ログファイルをクリアする
    NOT_CLEAR_LOG = 2  # ログファイルをクリアしない


class OpenParamExchange(Enum):
    """Open関数のパラメータ（データ変換）"""

    CONVERT_AND_OPEN = 1  # 変換して開く
    COPY_CONVERT_AND_OPEN = 2  # コピーして変換して開く
    NOT_OPEN = 3  # 開かない


class OpenParamBuckup(Enum):
    """Open関数のパラメータ（バックアップ）"""

    BACKUP_AND_OPEN = 1  # バックアップを別名保存して開く
    DELETE_BACKUP_AND_OPEN = 2  # バックアップは削除して開く
    NOT_OPEN = 3  # 開かない


class CloseParam(Enum):
    """Close関数のパラメータ"""

    SAVE_AND_CLOSE = 1  # 保存する
    NOT_SAVE_AND_CLOSE = 2  # 保存しない


class ExportResultCsvParamSave(Enum):
    """ExportResultCsv関数のパラメータ"""

    OVERWRITE = 1  # 上書きする
    ABORT = 2  # 中断する
    SAVE_AS = 3  # 別名で保存する


class ExportResultCsvParamSymbol(Enum):
    """ExportResultCsv関数のパラメータ"""

    OMIT_SYMBOL = 1  # 記号を省略する
    NOT_OMIT_SYMBOL = 2  # 記号を省略しない


class ExportResultCsvParamOutput(Enum):
    """ExportResultCsv関数のパラメータ"""

    ALL = 1  # 全部材出力
    REPRESENTATIVE = 2  # 代表部材出力


class ContainsKozo(Enum):
    """主体構造の有無"""

    NO = 0  # なし
    YES = 1  # あり
    CFT = 2  # CFT（S造用）


class Ss7PythonUtils:
    """
    『Op.Python実行』の共通モジュール
    『Op.Python実行』の各関数をラップし、エラーがある場合はエラー情報を返す。
    """

    def __init__(self, cmd: Cmd, version: Optional[str] = None):
        """
        『Op.Python実行』の初期化

        Args:
            cmd (Cmd): 『Op.Python実行』のインスタンス
            version (Optional[str], optional): 『SS7』のバージョン。Noneの場合、最新バージョンを使用する。
        """

        self.version = version
        self.cmd = cmd

        # 『Op.Python実行』の初期化
        self.cmd.Init()

    def start(self, startparam: StartParam) -> Optional[ErrInfo]:
        """
        『Op.Python実行』を開始する関数

        Args:
            startparam (StartParam): 開始パラメータ

        Returns:
            Optional[ErrInfo]: エラー情報（ErrInfo）
        """
        self.cmd.Start(self.version, startparam.value)
        errinfo = self.cmd.GetLastError()
        if self.check_error(errinfo):
            return None
        else:
            return errinfo

    def open(
        self,
        ikn_folder: str,
        openparam_exchange: OpenParamExchange,
        openparam_backup: OpenParamBuckup,
    ) -> Union[Ss7Data, ErrInfo]:
        """
        物件データを開く関数

        Args:
            ikn_folder (str): 物件データのフォルダパス（※絶対パス）
            openparam_exchange (OpenParamExchange): データ変換のパラメータ
            openparam_backup (OpenParamBuckup): バックアップのパラメータ

        Returns:
            Union[Ss7Data, ErrInfo]: 物件データ（Ss7Data）またはエラー情報（ErrInfo）
        """
        data: Ss7Data = self.cmd.Open(
            ikn_folder, openparam_exchange.value, openparam_backup.value
        )
        errinfo = self.cmd.GetLastError()
        if self.check_error(errinfo) and data is not None:
            return data
        else:
            return errinfo

    def save(self, data: Ss7Data) -> Optional[ErrInfo]:
        """
        物件データを保存する関数

        Args:
            data (Ss7Data): 物件データ

        Returns:
            Optional[ErrInfo]: エラー情報（ErrInfo）
        """
        data.Save()
        errinfo = self.cmd.GetLastError()
        if self.check_error(errinfo):
            return None
        else:
            return errinfo

    def close(self, data: Ss7Data, close_param: CloseParam) -> Optional[ErrInfo]:
        """
        物件データを閉じる関数

        Args:
            data (Ss7Data): 物件データ
            close_param (CloseParam): 閉じるパラメータ

        Returns:
            Optional[ErrInfo]: エラー情報（ErrInfo）
        """
        data.Close(close_param.value)
        errinfo = self.cmd.GetLastError()
        if self.check_error(errinfo):
            return None
        else:
            return errinfo

    def get_info_floor(self, ikn_folder: str) -> Union[int, ErrInfo]:
        """
        階数を取得する関数

        Args:
            ikn_path (str): 物件データのフォルダパス（※絶対パス）

        Returns:
            Union[int, ErrInfo]: 階数またはエラー情報（ErrInfo）
        """
        floor = self.cmd.GetInfoFloor(ikn_folder)
        errinfo = self.cmd.GetLastError()
        if self.check_error(errinfo):
            return floor
        else:
            return errinfo

    def get_info_span_x(self, ikn_folder: str) -> Union[float, ErrInfo]:
        """
        X方向のスパンを取得する関数

        Args:
            ikn_path (str): 物件データのフォルダパス（※絶対パス）

        Returns:
            Union[float, ErrInfo]: X方向のスパンまたはエラー情報（ErrInfo）
        """
        spanx = self.cmd.GetInfoSpanX(ikn_folder)
        errinfo = self.cmd.GetLastError()
        if self.check_error(errinfo):
            return spanx
        else:
            return errinfo

    def get_info_span_y(self, ikn_folder: str) -> Union[float, ErrInfo]:
        """
        Y方向のスパンを取得する関数

        Args:
            ikn_path (str): 物件データのフォルダパス（※絶対パス）

        Returns:
            Union[float, ErrInfo]: Y方向のスパンまたはエラー情報（ErrInfo）
        """
        spany = self.cmd.GetInfoSpanY(ikn_folder)
        errinfo = self.cmd.GetLastError()
        if self.check_error(errinfo):
            return spany
        else:
            return errinfo

    def get_info_kozo_rc(self, ikn_folder: str) -> Union[bool, ErrInfo]:
        """
        主体構造にRCが含まれるか判別する関数

        Args:
            ikn_path (str): 物件データのフォルダパス（※絶対パス）

        Returns:
            Union[bool, ErrInfo]: 主体構造にRCが含まれるか判別結果またはエラー情報（ErrInfo）
        """
        kozo_rc = self.cmd.GetInfoKozoRC(ikn_folder)
        errinfo = self.cmd.GetLastError()
        if self.check_error(errinfo):
            return kozo_rc == ContainsKozo.YES.value
        else:
            return errinfo

    def get_info_kozo_src(self, ikn_folder: str) -> Union[bool, ErrInfo]:
        """
        主体構造にSRCが含まれるか判別する関数

        Args:
            ikn_path (str): 物件データのフォルダパス（※絶対パス）

        Returns:
            Union[bool, ErrInfo]: 主体構造にSRCが含まれるか判別結果またはエラー情報（ErrInfo）
        """
        kozo_src = self.cmd.GetInfoKozoSRC(ikn_folder)
        errinfo = self.cmd.GetLastError()
        if self.check_error(errinfo):
            return kozo_src == ContainsKozo.YES.value
        else:
            return errinfo

    def get_info_kozo_s(self, ikn_folder: str) -> Union[bool, ErrInfo]:
        """
        主体構造にS造が含まれるか判別する関数

        Args:
            ikn_path (str): 物件データのフォルダパス（※絶対パス）

        Returns:
            Union[bool, ErrInfo]: 主体構造にS造が含まれるか判別結果またはエラー情報（ErrInfo）
        """
        kozo_s = self.cmd.GetInfoKozoS(ikn_folder)
        errinfo = self.cmd.GetLastError()
        if self.check_error(errinfo):
            return kozo_s == ContainsKozo.YES.value or kozo_s == ContainsKozo.CFT.value
        else:
            return errinfo

    def get_info_kozo_cft(self, ikn_folder: str) -> Union[bool, ErrInfo]:
        """
        主体構造にCFTが含まれるか判別する関数

        Args:
            ikn_path (str): 物件データのフォルダパス（※絶対パス）

        Returns:
            Union[bool, ErrInfo]: 主体構造にCFTが含まれるか判別結果またはエラー情報（ErrInfo）
        """
        kozo_cft = self.cmd.GetInfoKozoS(ikn_folder)
        errinfo = self.cmd.GetLastError()
        if self.check_error(errinfo):
            return kozo_cft == ContainsKozo.CFT.value
        else:
            return errinfo

    def is_calc_end(
        self, data: Ss7Data, result_no: str, calc_items: str
    ) -> Union[bool, ErrInfo]:
        """
        計算項目が計算済みかどうかを確認する関数

        Args:
            data (Ss7Data): 物件データ
            result_no (str): 結果番号(例:「結果1」)
            calc_items (list): 計算項目

        Returns:
            Union[bool, ErrInfo]: 計算済みの場合はTrue、エラーの場合はエラー情報（ErrInfo）
        """
        is_calc_end = data.IsCalcEnd(result_no, calc_items)
        errinfo = self.cmd.GetLastError()
        if self.check_error(errinfo):
            return is_calc_end
        else:
            return errinfo

    def calculate(
        self, data: Ss7Data, result_no: str, calc_items: str
    ) -> Optional[ErrInfo]:
        """
        計算を実行する関数

        Args:
            data (Ss7Data): 物件データ
            result_no (str): 結果番号(例:「結果1」)
            calc_items (list): 計算項目

        Returns:
            Optional[ErrInfo]: エラー情報（ErrInfo）
        """
        data.Calculate(result_no, calc_items)
        errinfo = self.cmd.GetLastError()
        if self.check_error(errinfo):
            return None
        else:
            return errinfo

    def get_result(self, data: Ss7Data, result_no: str) -> Union[Ss7Result, ErrInfo]:
        """
        結果データを取得する関数

        Args:
            data (Ss7Data): 物件データ
            result_no (str): 結果番号(例:「結果1」)

        Returns:
            Union[Ss7Result, ErrInfo]: 結果データ（Ss7Result）またはエラー情報（ErrInfo）
        """
        result: Ss7Result = data.GetResultData(result_no)
        errinfo = self.cmd.GetLastError()
        if self.check_error(errinfo) and result is not None:
            return result
        else:
            return errinfo

    def export_result_csv(
        self,
        result: Ss7Result,
        out_items: str,
        csv_path: str,
        save_param: ExportResultCsvParamSave,
        symbol_param: ExportResultCsvParamSymbol,
        output_param: ExportResultCsvParamOutput,
    ) -> Optional[ErrInfo]:
        """
        結果CSVを出力する関数

        Args:
            result (Ss7Result): 結果データ
            out_items (str): 出力項目
            csv_path (str): CSVファイルのパス
            save_param (ExportResultCsvParamSave): 保存パラメータ
            symbol_param (ExportResultCsvParamSymbol): 記号パラメータ
            output_param (ExportResultCsvParamOutput): 出力パラメータ

        Returns:
            Optional[ErrInfo]: エラー情報（ErrInfo）
        """
        result.ExportResultCsv(
            out_items,
            csv_path,
            save_param.value,
            symbol_param.value,
            output_param.value,
        )
        errinfo = self.cmd.GetLastError()
        if self.check_error(errinfo):
            return None
        else:
            return errinfo

    def check_error(self, errinfo: ErrInfo) -> bool:
        """
        『Op.Python実行』のエラーをチェックする関数

        Args:
            errinfo (ErrInfo): エラー情報

        Returns:
            bool: エラーがある場合はFalse、ない場合はTrue

        Note:
            直前の処理についてエラーの有無をチェックします。
        """
        if errinfo.IsOK():
            return True
        else:
            return False

    def get_error_no(self, errinfo: ErrInfo) -> int:
        """
        エラー番号を取得する関数

        Args:
            errinfo (ErrInfo): エラー情報

        Returns:
            int: エラー番号
        """
        return errinfo.GetErrorNo()

    def get_error_msg(self, errinfo: ErrInfo) -> str:
        """
        エラーメッセージを取得する関数

        Args:
            errinfo (ErrInfo): エラー情報

        Returns:
            str: エラーメッセージ
        """
        return errinfo.GetErrorMessage()
